/***************************************************************************
                        get_coeffs.c  -  description
                             -------------------
   Create transform coefficients using the exponents, and mantissas read 
    from the bitstream.  Rematrixing, and downmixing is also handled here.
                             -------------------
    begin                : Wed Oct 19 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>
#include <stddef.h>

#include "get_coeffs.h"

static const u8 REMATRIX_BAND_SIZE[4] = {25-13, 37-25, 61-37, AC3_MAX_CH_BIN-61};

void get_coeffs(AUDIO_BLOCK *blk)
{
   u8        *lfe_ptr = (u8 *)blk->lfe_ptr,
             *exp_ptr = &(blk->ch_data[0].num_bins);
   AC3_COEFF *coeff   =   blk->ch_data[0].coeff;
   AC3_COEFF  coeff_sum, coeff_diff;
   u32        num_bins, cnt;
   u8         rematrix, cpl_bins;

   /* shift mantissas by their respective exponent values */
   do {
      num_bins = cnt = (u32)*exp_ptr;
      exp_ptr += offsetof(CH_STRUCT, exp);
      while (cnt--) {
#if (AC3_COEFF_SIZE > 23)
         *coeff >>= (*exp_ptr + 1); /* - *gainrng;*/
#else
         *coeff >>= (*exp_ptr > (AC3_COEFF_SIZE-2)) ?
                     (AC3_COEFF_SIZE-1) : (*exp_ptr + 1);
#endif
         exp_ptr++;
         coeff++;
      }
      exp_ptr += (sizeof(CH_STRUCT)-offsetof(CH_STRUCT, exp)) - num_bins;
      coeff   += (sizeof(CH_STRUCT)/sizeof(AC3_COEFF))        - num_bins;
   } while (exp_ptr < lfe_ptr);
#if (AC3_OUTPUT_CHANNELS > 2)
   if (*lfe_ptr) {
      cnt = AC3_MAX_LFE_BIN - 1;
      do {
         blk->lfe_ptr->coeff[cnt] >>= blk->lfe_ptr->exp[cnt];
      } while (cnt--);
   }
#endif

   /* undo the rematrixing, if present */
   if (blk->rematrix) {
#define RGT       (sizeof(CH_STRUCT)/sizeof(AC3_COEFF))
#define RBND_SIZE exp_ptr
      rematrix = blk->rematrix << 1;
      if (blk->ch_data[0].num_bins > blk->ch_data[1].num_bins)
         num_bins = (u32)blk->ch_data[1].num_bins;
      else
         num_bins = (u32)blk->ch_data[0].num_bins;
      if (num_bins > 13) {
         num_bins -= 13;
         coeff    -= (2*(sizeof(CH_STRUCT)/sizeof(AC3_COEFF))) - 13;
         RBND_SIZE = (u8 *)REMATRIX_BAND_SIZE;
         do {
            cnt = (u32)*(RBND_SIZE++);
            if (rematrix >= 0x80) {
               do {
                  coeff_sum    = *coeff;
                  coeff_diff   = *(coeff+RGT);
                  *coeff       = coeff_sum + coeff_diff;
                  *(coeff+RGT) = coeff_sum - coeff_diff;
                  coeff++;
               } while ( (--num_bins)&&(--cnt) );
            }
            else if (num_bins <= cnt)
               break;
            else {
               coeff    += cnt;
               num_bins -= cnt;
            }
            rematrix <<= 1;
         } while ( (num_bins)&&(rematrix) );
      }
#undef RGT
#undef RBND_SIZE
   }

   /* make sure the high coefficients, that weren't in the 
       bitstream, are zero 
      downmix to 2 channels, if necessary */
#define CH_BINS   exp_ptr
#define DOWNMIX   rematrix
#define PREV      (sizeof(CH_STRUCT)/sizeof(AC3_COEFF))
#if (AC3_OUTPUT_CHANNELS == 2)
   cnt      = 0;
   DOWNMIX  = blk->ch;
#endif
   cpl_bins = blk->cpl_data.num_bins;
   CH_BINS  = &(blk->ch_data[0].num_bins);
   coeff    =   blk->ch_data[0].coeff;
   do {
      num_bins = (u32)*CH_BINS;
      if ( (cpl_bins)&&(*(CH_BINS+offsetof(CH_STRUCT, cpl_coord))) )
         num_bins += (u32)cpl_bins;    /* add coupling bins */
      if (num_bins > AC3_MAX_CH_BIN)
         num_bins = AC3_MAX_CH_BIN;
      memset((coeff+num_bins), 0, 
             (size_t)(AC3_NUM_BINS-num_bins)*sizeof(AC3_COEFF));
#if (AC3_OUTPUT_CHANNELS == 2)
      if ( (DOWNMIX)&&(cnt) ) {
         coeff += (cnt < num_bins) ? num_bins : cnt;
         do {
            coeff--;
            *(coeff-PREV) >>= 1;
            *(coeff-PREV)  += *coeff >> 1;
         } while (--cnt);
      }
      cnt = num_bins;
#endif
      CH_BINS += sizeof(CH_STRUCT);
      coeff   += sizeof(CH_STRUCT)/sizeof(AC3_COEFF);
   } while (CH_BINS < lfe_ptr);
#if (AC3_OUTPUT_CHANNELS == 2)
   if (DOWNMIX)
      blk->lfe_ptr = (LFE_STRUCT *)&(blk->ch_data[2]);
#endif
#undef CH_BINS
#undef DOWNMIX
#undef PREV
}

