/***************************************************************************
                          PAT_module.c  -  description
                             -------------------
   PROGRAM ASSOCIATION TABLE module functions.
                             -------------------
    begin                : Sat Aug 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/* #include <stdio.h> */

#include "PSI_SECTION.h"
#include "PAT_module.h"

#ifndef PAT_MAX_ENTRIES
# define PAT_MAX_ENTRIES 8
#endif

/*int PAT_MODULE_PRINT_DATA = 0;*/


/* PRIVATE STRUCTURES */
typedef struct {
   u16 program_number,          /* 16: Specifies the program to which the
                                         given program_map_PID is valid. All
                                         non-zero values are user defined.
                                         This value must be unique in all
                                         versions of the table (each
                                         program_number can only point to one
                                         program_map_PID). */
       program_map_PID;         /* 13: The PID of the TS packets carrying the
                                         Program Map Table. This PID can be
                                         shared among many program_numbers. */
} PAT_ENTRY;

typedef struct {
   u8        version, nex_sec, size, active;
   PAT_ENTRY entry[PAT_MAX_ENTRIES];
} PAT;

typedef struct {
   u16 entry_added, table_id_ext;
   PAT tbl1, tbl2;
} PAT_module_data;

/* LOCAL FUNCTIONS */
static void PAT_ENTRY_read(const u8  *data_stream,
                           u32        num,
                           PAT_ENTRY *pat);
/*
static void PAT_display(PAT *tbl, u32 table_id_ext);
*/                          

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PAT_PSI_MODULE_mem_needed()
{
   return sizeof(PAT_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void PAT_PSI_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   PAT_PSI_MODULE_done(mem);
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void PAT_PSI_MODULE_done(void *mem)
{
   PAT_module_data *data = (PAT_module_data *)mem;

   data->entry_added  = data->table_id_ext = 0;
   data->tbl1.version = data->tbl1.nex_sec = 0;
   data->tbl1.size    = data->tbl1.active  = 0;
   data->tbl2.version = data->tbl2.nex_sec = 0;
   data->tbl2.size    = data->tbl2.active  = 0;
}

/**************************************************************************
   COMMON PSI MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the PSI_SECTIONs
       they produce.
 **************************************************************************/
/* PSI modules should always return 0. */
int PAT_PSI_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx)
{
   return 0;
}

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PSI_SECTION that needs handling. */
void PAT_PSI_MODULE_process_pkt(void *mem, void *psi)
{
   PAT             *tbl;
   PAT_module_data *data    = (PAT_module_data *)mem;
   PSI_SECTION     *psi_sec = (PSI_SECTION *)psi;
   u8               psi_ver = PSI_SECTION_VERSION_NUMBER(psi_sec);

   if ( (data->entry_added)||
        (psi_sec->table_id != PSI_SECTION_PROGRAM_ASSOCIATION_ID) )
      return;  /* PSI_SECTION does not contain PAT data */
   else if ((u16)data->table_id_ext != psi_sec->table_id_extension) {
      /* id extension value does not match, so clear all tables */
      PAT_PSI_MODULE_init(data, NULL);
      data->table_id_ext = (u32)psi_sec->table_id_extension;
   }
      
   /* find table for this PSI_SECTION */
   if ( (!data->tbl1.nex_sec)||(data->tbl1.version == psi_ver) )
      tbl = &(data->tbl1);
   else if ( (!data->tbl2.nex_sec)||(data->tbl2.version == psi_ver) )
      tbl = &(data->tbl2);
   else {
      /* overwrite table with new data */
      if (data->tbl1.active)
         tbl = &(data->tbl2);
      else
         tbl = &(data->tbl1);
      tbl->nex_sec = tbl->size = tbl->active = 0;
   }

   if (tbl->nex_sec == psi_sec->section_number) {
      if (!tbl->nex_sec)
         tbl->version = psi_ver;    /* store the version number */

      if (tbl->size < PAT_MAX_ENTRIES) {
         psi_sec->payload_len16 >>= 2;
         if ((psi_sec->payload_len16 + (u16)tbl->size) > PAT_MAX_ENTRIES)
            psi_sec->payload_len16 = PAT_MAX_ENTRIES - (u16)tbl->size;
         
         /* read new entries into the table */
         PAT_ENTRY_read(psi_sec->payload,
                        (u32)psi_sec->payload_len16,
                        (tbl->entry + tbl->size));
         tbl->size += (u8)psi_sec->payload_len16;
      }
      tbl->nex_sec++;
   }

   if (psi_sec->section_number == psi_sec->last_section_number) {
      /* check if table needs to be activated */
      if (PSI_SECTION_CURRENT_NEXT_FLAG(psi_sec)) {
         if (!tbl->active) {
            tbl->active = 1;
            if (tbl == &(data->tbl1))
               tbl = &(data->tbl2);
            else
               tbl = &(data->tbl1);
            tbl->active = 0;
         }
      }
      else if (tbl->active)
         tbl->active = 0;
/*
      if (PAT_MODULE_PRINT_DATA) {
         if (tbl->version != psi_ver) {
            if (tbl == &(data->tbl1))
               tbl = &(data->tbl2);
            else
               tbl = &(data->tbl1);
         }
         
         if (tbl->nex_sec == (psi_sec->section_number+1)) {
            tbl->nex_sec++;
            PAT_display(tbl, data->table_id_ext);
         }
      }
*/
   }
}


/*************************************************************************
   PAT MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to get the PAT data.
       The PAT data is extracted from the PSI_SECTIONs sent by a system
       module.
 *************************************************************************/
/* Adds an entry to the PAT, using the given values.  This function is
     used when there is no PAT data in the stream.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number of the new entry. Any value greater
                        than 0 is valid.
             u16    - Program Map PID of the new entry. Any unused user
                        defined PID value is valid. */
void PAT_PSI_MODULE_add_entry(void *mem, u16 program_number, u16 PMT_PID)
{
   PAT             *cur_tbl;
   PAT_module_data *data = (PAT_module_data *)mem;

   if (data->tbl2.active)
      cur_tbl = &(data->tbl2);
   else
      cur_tbl = &(data->tbl1);

   data->entry_added  = 1;
   data->table_id_ext = 0;
   
   cur_tbl->version = cur_tbl->nex_sec = 0xFF;
   cur_tbl->size    = cur_tbl->active  =    1;
   
   cur_tbl->entry[0].program_number  = program_number;
   cur_tbl->entry[0].program_map_PID = PMT_PID;
/*
   if (PAT_MODULE_PRINT_DATA)
      PAT_display(cur_tbl, data->table_id_ext);
*/
}

/* Returns the PID of the TS_PKTs that contains the PMT data of the
     specified program number.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Program number to search for.
      RETURN: Program found     - Valid user defined PID value.
              Program not found - PID value that is >= TS_PKT_NUL_ID. */
u16 PAT_PSI_MODULE_get_pid(void *mem, u16 program_number)
{
   PAT             *cur_tbl;
   PAT_module_data *data = (PAT_module_data *)mem;

   if (data->tbl1.active)
      cur_tbl = &(data->tbl1);
   else if (data->tbl2.active)
      cur_tbl = &(data->tbl2);
   else
      program_number = 0;

   if (program_number) {      
      u8 i;
      
      for (i = 0; i < cur_tbl->size; i++) {
         if (cur_tbl->entry[i].program_number == program_number)
            return cur_tbl->entry[i].program_map_PID;
      }
   }
   return 0xFFFF;
}

/* Returns the program number at the specified location, in the
     access table.
      INPUT: void * - Pointer to the memory provided for the module.
             u16    - Table index to search for.
      RETURN: Program found     - Valid program number.
              Program not found - 0 */
u16 PAT_PSI_MODULE_get_program(void *mem, u16 idx)
{
   PAT             *cur_tbl;
   PAT_module_data *data = (PAT_module_data *)mem;

   if (data->tbl1.active)
      cur_tbl = &(data->tbl1);
   else if (data->tbl2.active)
      cur_tbl = &(data->tbl2);
   else
      idx = PAT_MAX_ENTRIES;

   if ( (idx < PAT_MAX_ENTRIES)&&(idx < (u16)cur_tbl->size) )
      return cur_tbl->entry[idx].program_number;
   return 0;
}


static void PAT_ENTRY_read(const u8  *data_stream,
                           u32        num,
                           PAT_ENTRY *pat)
{
   while (num) {
      pat->program_number  = (u16)(*data_stream) << 8;
      data_stream++;

      pat->program_number |= (u16)(*data_stream);
      data_stream++;

      pat->program_map_PID  = (u16)(*data_stream & 0x1F) << 8;
      data_stream++;

      pat->program_map_PID |= (u16)(*data_stream);
      data_stream++;
      pat++;
      num--;
   }
}

/*
static void PAT_display(PAT *tbl, u32 table_id_ext)
{
   u8 i;
   
   printf("\nProgram Access Table\n"
          "   table_id_extension = 0x%04X\n"
          "   version            = %6u\n"
          "   active             =      %c\n"
          "     PROGRAM  PMT_PID\n", (u32)table_id_ext,
                                     (u32)tbl->version,
                                     ((tbl->active)?'Y':'N'));
   for (i = 0; i < tbl->size; i++) {
      printf("      0x%04hX   0x%04hX\n", tbl->entry[i].program_number,
                                          tbl->entry[i].program_map_PID);
   }
}
*/
#undef PAT_MAX_ENTRIES

