/***************************************************************************
                          PSI_module.c  -  description
                             -------------------
   PSI SYSTEM module functions.
                             -------------------
    begin                : Sat Aug 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "PSI_SECTION.h"
#include "PSI_module.h"

/* PRIVATE STRUCTURES */
typedef struct {
   MPEG2_PSI_PES_MODULE *psi_module;
   PSI_SECTION           psi;
   u8                   *pos_ptr;
   u8                    buf[PSI_SECTION_MAX_SIZE];
} PSI_module_data;


/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PSI_SYS_MODULE_mem_needed()
{
   return sizeof(PSI_module_data);
}

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void PSI_SYS_MODULE_init(void *mem, SYS_DATA *sys_data)
{
   PSI_module_data *data = (PSI_module_data *)mem;

   data->pos_ptr = NULL;   /* wait for next start indicator */
   data->psi_module->common.init(data->psi_module->common.mem, sys_data);
}

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void PSI_SYS_MODULE_done(void *mem)
{
   PSI_module_data *data = (PSI_module_data *)mem;

   data->pos_ptr = NULL;   /* wait for next start indicator */
   data->psi_module->common.done(data->psi_module->common.mem);
}

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void PSI_SYS_MODULE_discontinuity(void *mem)
{
   PSI_module_data *data = (PSI_module_data *)mem;

   data->pos_ptr = NULL;   /* wait for next start indicator */
   if (data->psi_module->common.discontinuity)
      data->psi_module->common.discontinuity(data->psi_module->common.mem);
}


/**************************************************************************
   SYSTEM MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to extract the PSI
       data from the corresponding TS_PKTs, and create PSI_SECTIONs for
       the PSI mdules.
 **************************************************************************/
/* Provides the PSI module that will handle the PSI_SECTIONs.
      INPUT: void *                 - Pointer to the memory provided for
                                        the module.
             MPEG2_PSI_PES_MODULE * - Pointer to a PSI module. */
void PSI_SYS_MODULE_psi_pes_module(void *mem, MPEG2_PSI_PES_MODULE *psi)
{
   PSI_module_data *data = (PSI_module_data *)mem;

   data->psi_module = psi;
}

/* Processes the packet handled by the module.
      INPUT: void *   - Pointer to the memory provided for the module.
             TS_PKT * - Pointer to packet that needs handling. */
void PSI_SYS_MODULE_process_pkt(void *mem, TS_PKT *ts)
{
   u32              offset, bytes_read;
   PSI_module_data *data = (PSI_module_data *)mem;

   if ( (!ts->payload)||(!ts->payload_len32) )
      return;  /* no data in this packet */
      
   if (!data->pos_ptr) {
      if (!ts->payload_unit_start_indicator)
         return;  /* waiting for start indicator, so ignore this packet */
      offset        = (u32)*(ts->payload) + 1;
      bytes_read    = 0;
      data->pos_ptr = data->buf;
   }
   else if (ts->payload_unit_start_indicator) {
      /* reached the end of a PSI_SECTION */
      offset     = (u32)*(ts->payload);
      bytes_read = (u32)(data->pos_ptr - data->buf);
      if ((offset+bytes_read) > PSI_SECTION_MAX_SIZE) {
         /* PSI_SECTION is not good, so consider it a discontinuity */
PSI_SYS_MODULE_process_pkt_DISCONTINUITY_1:
         data->psi_module->common.discontinuity(data->psi_module->common.mem);
      }
      else {
         if (offset) {
            memcpy(data->pos_ptr, (ts->payload+1), (size_t)offset);

            /* Read the PSI_SECTION, once we have at least 8 bytes */
            if ( (bytes_read < 8)&&((offset+bytes_read) >= 8)&&
                 (!PSI_SECTION_read(data->buf, &(data->psi))) )
               goto PSI_SYS_MODULE_process_pkt_DISCONTINUITY_1;
         }

         /* The entire PSI_SECTION is in the buffer, so process it */
         data->psi_module->process_pkt(data->psi_module->common.mem,
                                       &(data->psi));
      }
      offset++;
      bytes_read    = 0;
      data->pos_ptr = data->buf;
   }
   else {
      offset     = 0;
      bytes_read = (u32)(data->pos_ptr - data->buf);
   }

   if (ts->payload_len32 > offset) {
      ts->payload_len32 -= offset;
      if ((ts->payload_len32+bytes_read) > PSI_SECTION_MAX_SIZE) {
         /* PSI_SECTION is not good, so consider it a discontinuity */
PSI_SYS_MODULE_process_pkt_DISCONTINUITY_2:
         data->psi_module->common.discontinuity(data->psi_module->common.mem);
         data->pos_ptr = NULL;   /* wait for next start indicator */
      }
      else {
         memcpy(data->pos_ptr,
                (ts->payload+offset),
                (size_t)ts->payload_len32);
         data->pos_ptr += ts->payload_len32;

         /* Read the PSI_SECTION, once we have at least 8 bytes */
         if ( (bytes_read < 8)&&((ts->payload_len32+bytes_read) >= 8)&&
              (!PSI_SECTION_read(data->buf, &(data->psi))) )
            goto PSI_SYS_MODULE_process_pkt_DISCONTINUITY_2;
      }
   }
}
