/***************************************************************************
                          TS_PKT.c  -  description
                             -------------------
   The functions needed to deal with the transport stream packets.
                             -------------------
    begin                : Sun Jul 18 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>

#include "TS_PKT.h"

/* Tries to find the first TS_PKT in data_stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream. There should be at
                             least (TS_PKT_SIZE + 1) bytes to verify the
                             presence of a TS_PKT.
      RETURN: TS_PKT found     - A pointer to the TS_PKT_SYNC_BYTE, of the
                                   first TS_PKT.
              TS_PKT not found - NULL */
const u8 *TS_PKT_find(const u8 *data_stream, u32 data_len)
{
   do {
      /* look for the sync byte */
      while ( (--data_len)&&(*data_stream != TS_PKT_SYNC_BYTE) ) {
         data_stream++;
      }
      if (!data_len)
         return NULL;   /* ran out of data to search through */
   } while ( (data_len >= (TS_PKT_SIZE-1))&&
             (*(data_stream+TS_PKT_SIZE) != TS_PKT_SYNC_BYTE) );
   return data_stream;
}

/* Reads the first TS_PKT, in data_stream, into ts.
      INPUT: data_stream - Points to the first byte, after the
                             TS_PKT_SYNC_BYTE, in the TS_PKT. This function
                             assumes there are at least (TS_PKT_SIZE - 1)
                             bytes in data_stream.
      OUTPUT: ts - Points to the TS_PKT that will hold data taken from
                     data_stream.
      RETURN: Success - A pointer to the TS_PKT_SYNC_BYTE of the next TS_PKT
                          in data_stream (you must verify this yourself!).
              Error   - NULL */
const u8 *TS_PKT_read(const u8 *data_stream, TS_PKT *ts)
{
   ts->transport_error_indicator    = *data_stream & 0x80;
   ts->payload_unit_start_indicator = *data_stream & 0x40;
   ts->transport_priority           = *data_stream & 0x20;
   ts->PID                          = (u16)(*data_stream & 0x1F) << 8;
   
   data_stream++;
   ts->PID                         |= (u16)(*data_stream);

   data_stream++;
   ts->transport_scrambling_control = *data_stream & 0xC0;
   ts->adaptation_field_control     = *data_stream & 0x30;
   ts->continuity_counter           = *data_stream & 0x0F;

   data_stream++;
   if (ts->adaptation_field_control & TS_PKT_HAS_ADAPT_FIELD) {
      u32 af_len = (u32)(*data_stream);

      ts->adapt_field_pos = (af_len) ? data_stream : NULL;
      
      if (ts->adaptation_field_control == TS_PKT_HAS_ADAPT_FIELD) {
         if (af_len > 183)
            return NULL;   /* corrupted packet */
         ts->payload       = NULL;
         ts->payload_len32 = 0;
      }
      else if (af_len > 182)
         return NULL;      /* corrupted packet */
      else {
         af_len++;
         ts->payload       = data_stream     + af_len;
         ts->payload_len32 = (TS_PKT_SIZE-4) - af_len;
      }
   }
   else {
      ts->adapt_field_pos = NULL;
      
      if (ts->adaptation_field_control) {
         ts->payload       = data_stream;
         ts->payload_len32 = TS_PKT_SIZE-4;
      }
      else {
         ts->payload       = NULL;
         ts->payload_len32 = 0;
      }
   }
   return (data_stream+TS_PKT_SIZE-4);
}
