/***************************************************************************
                        MPEG_audio.c  -  description
                             -------------------
   Functions for handling mpeg audio frames.
                             -------------------
    begin                : Thu Apr 7 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define MPEG_AUDIO_TRACE*/
#ifdef MPEG_AUDIO_TRACE
# include <stdio.h>
#endif
#include <string.h>

#include "MPEG_audio.h"

static const u16 bitrate[]    = {  0,  32,  48,  56,
                                  64,  80,  96, 112,
                                 128, 160, 192, 224,
                                 256, 320, 384,   1},
                 samplerate[] = {441, 480, 320,   1};

static u32 find_sblimit(u32 sample_rate, u32 frame_len, u32 num_ch);


/* Find the first MPEG audio frame in the given data stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream.
             frame       - Pointer to an MPEG_AUDIO_FRAME.
      RETURN: Success         - Returns number of bytes read from
                                  data_stream.
              Frame not found - 0 */
u32 MPEG_AUDIO_next_frame(const u8         *data_stream,
                          u32               data_len,
                          MPEG_AUDIO_FRAME *frame)
{
   const u8 *tmp,
            *ptr = data_stream;
   
   while (data_len) {
      switch (frame->state) {
      case MPEG_AUDIO_next_frame_STATE+0:
         tmp = memchr(ptr, MPEG_AUDIO_SYNC_BYTE, (size_t)data_len);
         if (!tmp)
            data_len  = 1;
         else {
            data_len -= (u32)(tmp - ptr);
            ptr       = tmp;
            frame->state++;
         }
         break;

      case MPEG_AUDIO_next_frame_STATE+1:
         frame->layer = *ptr;
         frame->state++;
#ifdef MPEG_AUDIO_TRACE
         printf("MPEG_AUDIO_FRAME: ");
         if (*ptr >= (MPEG_AUDIO_LAYER1_BYTE-1))
            printf("LAYER 1\n");
         else if (*ptr >= (MPEG_AUDIO_LAYER2_BYTE-1))
            printf("LAYER 2\n");
         else if (*ptr >= (MPEG_AUDIO_LAYER3_BYTE-1))
            printf("LAYER 3\n");
         else
            printf("UNKNOWN 0x%02hX\n", (u16)*ptr);
#endif
         break;

      case MPEG_AUDIO_next_frame_STATE+2:
         frame->sample_rate  = (u32)samplerate[(0x0C & *ptr) >> 2];
         if (*ptr >= 0x10) {
            frame->frame_len  = 1440 * (u32)bitrate[*ptr >> 4];
            frame->frame_len /= frame->sample_rate;
            if (*ptr & 0x02)
               frame->frame_len++;
         }
         else if (!frame->frame_len) {
            /* free bitrate, need to find next sync byte to calculate frame_len */
            if (!(*ptr & 0x02))
               frame->frame_len = 1;
            frame->state = MPEG_AUDIO_next_frame_STATE+5;
            break;
         }
         frame->sample_rate *= 100;
         frame->state++;
#ifdef MPEG_AUDIO_TRACE
         printf("  bitrate=%huKbps  samplerate=%uHz  frame_len=%ubytes\n",
                bitrate[*ptr >> 4], frame->sample_rate, frame->frame_len);
#endif
         break;

      case MPEG_AUDIO_next_frame_STATE+3:
         if (*ptr >= 0xC0)
            frame->num_ch = 1;
         else
            frame->num_ch = 2;
         frame->sblimit = (u8)find_sblimit(frame->sample_rate, 
                                           frame->frame_len,
                                           (u32)frame->num_ch);
         if ( (0x40 <= *ptr)&&(*ptr <= 0x7F) )  /* JOINT_STEREO */
            frame->bound = (0x0C & (*ptr >> 2)) + 4;
         else
            frame->bound = frame->sblimit;
         frame->state++;
#ifdef MPEG_AUDIO_TRACE
         printf("  mode=");
         if (*ptr >= 0xC0)
            printf("SINGLE_CHANNEL");
         else if (*ptr >= 0x80)
            printf("DUAL_CHANNEL");
         else if (*ptr >= 0x40)
            printf("JOINT_STEREO");
         else
            printf("STEREO");
         printf("  num_ch=%hu  bound=%hu  sblimit=%hu  crc=%s\n", 
                (u16)frame->num_ch,
                (u16)frame->bound,
                (u16)frame->sblimit,
                (frame->layer & 1) ? "no" : "yes");
#endif
         break;

      case MPEG_AUDIO_next_frame_STATE+4:
         if (frame->layer & 0x01) {
            frame->state         = MPEG_AUDIO_next_frame_END_STATE;
            frame->payload_len16 = (u16)data_len;
            frame->payload       = ptr;
            return (u32)(ptr - data_stream);
         }
         
         /* skip crc bytes */
         frame->state = MPEG_AUDIO_next_frame_STATE+7;
         break;

      case MPEG_AUDIO_next_frame_STATE+5:  /* find next sync byte */
         tmp = memchr(ptr, MPEG_AUDIO_SYNC_BYTE, (size_t)data_len);
         if (!tmp) {
            frame->frame_len += data_len;
            data_len          = 1;
         }
         else {
            frame->frame_len += (u32)(tmp - ptr);
            data_len         -= (u32)(tmp - ptr);
            ptr               = tmp;
            frame->state++;
         }
         break;
         
      case MPEG_AUDIO_next_frame_STATE+6:
         frame->frame_len++;  /* got the frame len */
         frame->state = (frame->layer == *ptr) ? 
                        MPEG_AUDIO_next_frame_STATE+2 :
                        MPEG_AUDIO_next_frame_STATE+5;
         break;
         
      case MPEG_AUDIO_next_frame_STATE+7:
         frame->layer     |= 0x01;  /* crc bytes have passed */
         frame->frame_len -= 2;
         frame->state      = MPEG_AUDIO_next_frame_STATE+4;
         break;

      default:
         frame->state = MPEG_AUDIO_next_frame_STATE+0;
/*         data_len     = 1;*/
      }
      ptr++;
      data_len--;
   }
   return 0;
}


static u32 find_sblimit(u32 sample_rate, u32 frame_len, u32 num_ch)
{
   frame_len >>= num_ch - 1;
   
   if (sample_rate == 48000) {
      if (frame_len >= 167)
         return 27;
      else if (frame_len <= 145)
         return  8;
   }
   else if (sample_rate == 44100) {
      if (frame_len >= 312)
         return 30;
      else if (frame_len >= 181)
         return 27;
      else if (frame_len <= 157)
         return  8;
   }
   else /*if (sample_rate == 32000)*/ {
      if (frame_len >= 431)
         return 30;
      else if (frame_len >= 251)
         return 27;
   }
   return 12;
}
